#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <unistd.h>
#include <signal.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <getopt.h>

#include "icmpecho.h"
#include "resolve.h"
#include "checksum.h"

#define MAX_TRANSFER_UNIT	1500
#define FRAGMENT_SIZE		MAX_TRANSFER_UNIT - sizeof(struct icmpdgramhdr)
#define DEFAULT_SIZE		64
#define DEFAULT_COUNT		1

#define MIN(X, Y)		((X) > (Y) ? (Y) : (X))

void quit();
void sigint(int sig);

static long c, t = 0;


int main(int argc, char *argv[])
{
	long count = DEFAULT_COUNT, size = DEFAULT_SIZE;
	struct sockaddr_in to;
	char buf[MAX_TRANSFER_UNIT];
	struct icmpdgramhdr *icmpecho = (struct icmpdgramhdr *)buf;
	struct iphdr *ip = &icmpecho->ip;
	struct icmphdr *icmp = &icmpecho->icmp;
	int ch, tot_len = MAX_TRANSFER_UNIT, raw_ip_socket, o_host = 0;

	while ((ch = getopt(argc, argv, "c:s:h:")) != EOF)
		switch ((char)ch) {
		case 'c':
			count = atol(optarg);
			break;
		case 's':
			size = atol(optarg);
			break;
		case 'h':
			ip->saddr = resolve(optarg);
			o_host = 1;
			break;
		}

	argv += optind;
	argc -= optind;

	if (argc != 1) {
		printf(
			"usage: yaping "
			"[-c count] "
			"[-s size] "
			"[-h host] "
			"host "
			"\n"
		);
		exit(1);
	}

	raw_ip_socket = socket(AF_INET, SOCK_RAW, IPPROTO_RAW);
	if (raw_ip_socket < 0) {
		perror("socket");
		exit(1);
	}	

	signal(SIGINT, quit);

	icmp->type = ICMP_ECHO;
	icmp->code = 0;
	icmp->un.echo.id = 0;
	icmp->un.echo.sequence = 0;
	icmp->checksum = 0;
	icmp->checksum = htons(~ICMP_ECHO);

	ip->ihl = (sizeof(struct iphdr) >> 2);
	ip->version = 4;
	ip->tot_len = htons(tot_len);
	ip->tos = 0;
	ip->frag_off = 0;
	ip->ttl = 64;
	ip->protocol = IPPROTO_ICMP;
	ip->daddr = resolve(argv[0]);

	to.sin_family = AF_INET;
	to.sin_addr.s_addr = ip->daddr;

	srandom(time(NULL));

	for (c = 0; c < count; c++) {
		long offset;

		ip->id = htons(c);
		if (!o_host)
			ip->saddr = random();

		for (offset = 0; offset < size || !offset; offset += FRAGMENT_SIZE) {
			const long togo = size - offset;

			int r;

			if (togo < FRAGMENT_SIZE) {
				tot_len = sizeof(struct icmpdgramhdr) + togo;
				ip->tot_len = htons(tot_len);
			}

			ip->frag_off = htons(offset >> 3);
			ip->check = 0;
			ip->check = htons(checksum((const ushort *)ip, tot_len));

			r = sendto(raw_ip_socket, icmpecho, tot_len, 0, (struct sockaddr *)&to, sizeof(to)); 
			if (r != tot_len)
				perror("sendto");
			else
				t += r;
		}
	}

	quit();
}

void quit()
{
	printf("%li icmp echo requests [%li octets]\n", c, t);

	exit(0);
}

void sigint(int sig)
{
	quit();
}
